# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from pyelasticsearch.exceptions import (
    ElasticHttpNotFoundError,
)
from charmworld.models import (
    CharmSource,
    QADataSource,
)
from charmworld.testing import (
    factory,
    MongoTestBase,
)
from charmworld.migrations import migrate
from charmworld.migrations.migrate import Versions

import mock


class MigrationTestBase(MongoTestBase):

    def setUp(self):
        super(MigrationTestBase, self).setUp()
        self.versions = Versions(migrate.get_migration_path())
        self.use_context(mock.patch.object(migrate, 'configure_logging'))


class TestMigration010(MigrationTestBase):

    def test_migration(self):
        self.use_index_client()
        source = CharmSource.from_request(self)
        source.save({
            '_id': 'a',
            'config': None,
        })
        source.save({
            '_id': 'b',
            'config': {
                'options': {},
            },
        })
        source.save({
            '_id': 'c',
            'config': {
                'options': {'foo': {'default': 'bar'}},
            },
        })
        source.save({'_id': 'd'})
        # This is not acceptable to Elasticsearch, because config is a list.
        self.db.charms.save({
            '_id': 'e',
            'config': ['a'],
        })
        self.versions.get_exodus('010_remap_options.py')(source, source)
        for charm in source._get_all('a'):
            self.assertEqual({
            '_id': 'a',
            'config': None,
        }, charm)
        for charm in source._get_all('b'):
            self.assertEqual({
                '_id': 'b',
                'config': {'options': []},
        }, charm)
        for charm in source._get_all('c'):
            self.assertEqual({
                '_id': 'c',
                'config': {'options': [{
                    'name': 'foo',
                    'default': 'bar',
                }]}
            }, charm)
        for charm in source._get_all('d'):
            self.assertEqual({'_id': 'd'}, charm)
        with self.assertRaises(ElasticHttpNotFoundError):
            self.index_client.get('e')
        self.assertEqual({'_id': 'e', 'config': ['a']},
                         self.db.charms.find_one({'_id': 'e'}))


class TestMigration011(MigrationTestBase):

    def test_migration(self):
        self.use_index_client()
        source = CharmSource.from_request(self)
        self.assertItemsEqual([], self.db.qa.find())
        self.versions.run_migration(self.db, self.index_client,
                                    '011_ensure_category_questions.py')
        self.assertIn('reliable', [cat['name'] for cat in self.db.qa.find()])


class TestMigration012(MigrationTestBase):

    def test_migration_updates_answers(self):
        self.use_index_client()
        charm1 = factory.get_charm_json()
        charm2 = factory.get_charm_json(name='name', series='series',
                                        owner='owner')
        charm2['qa'] = {'answer': 'data'}
        charm2['store_data']['revision'] = 5
        source = CharmSource.from_request(self)
        qa_data_source = QADataSource.from_db(self.db)
        source.save(charm1)
        source.save(charm2)
        self.assertItemsEqual([], qa_data_source.answers.find())
        self.versions.run_migration(self.db, self.index_client,
                                    '012_migrate_qa_representation.py')
        self.assertEqual([{
            '_id': '~owner/series/name/5',
            'name': 'name',
            'series': 'series',
            'owner': 'owner',
            'revision': 5,
            'answers': {'answer': 'data'}
        }], list(qa_data_source.answers.find()))

    def test_migration_removes_original_qa(self):
        self.use_index_client()
        source = CharmSource.from_request(self)
        charm = factory.get_charm_json()
        charm['qa'] = {'answer': 'data'}
        source.save(charm)
        for charm_data in source._get_all(charm['_id']):
            self.assertIn('qa', charm_data)
        self.versions.run_migration(self.db, self.index_client,
                                    '012_migrate_qa_representation.py')
        for charm_data in source._get_all(charm['_id']):
            self.assertNotIn('qa', charm_data)
