# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from mock import patch

from charmworld.health import (
    check_api,
    check_elasticsearch,
    check_mongodb,
)
from charmworld.testing import (
    factory,
    ViewTestBase,
)


class TestCheckMixin:

    def assertCheck(self, check, name, status, remark=None):
        self.assertEqual(name, check.name)
        self.assertEqual(status, check.status)
        if remark is not None:
            self.assertEqual(remark, check.remark)

    @staticmethod
    def makeFailCallable():
        def failing(*args, **kwargs):
            raise ValueError
        return failing


class HealthTestCase(ViewTestBase, TestCheckMixin):

    def test_check_mongodb_pass(self):
        factory.makeCharm(self.db)
        check = check_mongodb(self.getRequest())
        remark = '1 charms found'
        self.assertCheck(check, 'charms collection', 'Pass', remark)

    def test_check_mongodb_missing_data_fail(self):
        check = check_mongodb(self.getRequest())
        remark = 'There are no charms. Is ingest running?'
        self.assertCheck(check, 'charms collection', 'Fail', remark)

    def test_check_mongodb_no_collection_fail(self):
        request = self.getRequest()
        with patch.object(request, 'db', new_callable=self.makeFailCallable):
            check = check_mongodb(request)
        self.assertCheck(check, 'charms collection', 'Fail')

    def test_check_elasticsearch_pass(self):
        ignore, charm_data = factory.makeCharm(self.db)
        self.use_index_client()
        self.index_client.index_charm(charm_data)
        check = check_elasticsearch(self.getRequest())
        remark = '1 matches found'
        self.assertCheck(check, 'charms index', 'Pass', remark)

    def test_check_elasticsearch_missing_data_fail(self):
        self.use_index_client()
        request = self.getRequest()
        check = check_elasticsearch(request)
        remark = 'There are no matches. Is ingest running?'
        self.assertCheck(check, 'charms index', 'Fail', remark)

    def test_check_elasticsearch_no_index_fail(self):
        request = self.getRequest()
        # The index is not setup.
        check = check_elasticsearch(request)
        self.assertCheck(check, 'charms index', 'Fail')

    def test_check_api_pass(self):
        ignore, charm_data = factory.makeCharm(
            self.db, downloads=5, is_featured=True)
        self.use_index_client()
        self.index_client.index_charm(charm_data)
        check = check_api(self.getRequest())
        remark = 'Interesting has new, popular, and featured.'
        self.assertCheck(check, 'API interesting', 'Pass', remark)

    def test_check_api_missing_data_fail(self):
        ignore, charm_data = factory.makeCharm(self.db)
        self.use_index_client()
        self.index_client.index_charm(charm_data)
        # There are no featured or downloaded charms.
        check = check_api(self.getRequest())
        remark = 'Interesting is missing new, popular, or featured.'
        self.assertCheck(check, 'API interesting', 'Fail', remark)

    def test_check_api_no_matches_fail(self):
        request = self.getRequest()
        # The index is not setup for the interesting search.
        check = check_api(request)
        self.assertCheck(check, 'API interesting', 'Fail')
